"""
Infer ancestral sequences based on a tree.
"""

import os, shutil, time, json, sys
from Bio import Phylo
from .utils import write_json
from treetime.vcf_utils import read_vcf, write_vcf
from collections import defaultdict

def ancestral_sequence_inference(tree=None, aln=None, ref=None, infer_gtr=True,
                                 marginal=False):
    from treetime import TreeAnc
    tt = TreeAnc(tree=tree, aln=aln, ref=ref, gtr='JC69', verbose=1)

    # convert marginal (from args.inference) from 'joint' or 'marginal' to True or False
    bool_marginal = (marginal == "marginal")

    # only infer ancestral sequences, leave branch length untouched
    tt.infer_ancestral_sequences(infer_gtr=infer_gtr, marginal=bool_marginal)

    print("\nInferred ancestral sequence states using TreeTime:"
          "\n\tSagulenko et al. TreeTime: Maximum-likelihood phylodynamic analysis"
          "\n\tVirus Evolution, vol 4, https://academic.oup.com/ve/article/4/1/vex042/4794731\n")

    return tt

def collect_sequences_and_mutations(T, is_vcf=False):
    data = defaultdict(dict)
    inc = 1 # convert python numbering to start-at-1
    for n in T.find_clades():
        if hasattr(n, "mutations"):
            mutations_attr = n.__getattribute__("mutations")
            data[n.name]['muts'] = [str(a)+str(int(pos)+inc)+str(d)
                                    for a,pos,d in mutations_attr]
    if not is_vcf:
        for n in T.find_clades():
            if hasattr(n, "sequence"):
                sequence_attr = n.__getattribute__("sequence")
                data[n.name]['sequence'] = ''.join(sequence_attr)
            else:
                data[n.name]['sequence'] = ''

    return data


def register_arguments(parser):
    parser.add_argument('--tree', '-t', required=True, help="prebuilt Newick")
    parser.add_argument('--alignment', '-a', help="alignment in fasta or VCF format")
    parser.add_argument('--output', '-o', type=str, help='file name to save mutations and ancestral sequences to')
    parser.add_argument('--inference', default='joint', choices=["joint", "marginal"],
                                    help="calculate joint or marginal maximum likelihood ancestral sequence states")
    parser.add_argument('--vcf-reference', type=str, help='fasta file of the sequence the VCF was mapped to')
    parser.add_argument('--output-vcf', type=str, help='name of output VCF file which will include ancestral seqs')


def run(args):
    # check alignment type, set flags, read in if VCF
    is_vcf = False
    ref = None
    anc_seqs = {}
    # check if tree is provided and can be read
    for fmt in ["newick", "nexus"]:
        try:
            T = Phylo.read(args.tree, fmt)
            break
        except:
            pass
    if T is None:
        print("ERROR: reading tree from %s failed."%args.tree)
        return 1

    if any([args.alignment.lower().endswith(x) for x in ['.vcf', '.vcf.gz']]):
        if not args.vcf_reference:
            print("ERROR: a reference Fasta is required with VCF-format alignments")
            return 1

        compress_seq = read_vcf(args.alignment, args.vcf_reference)
        aln = compress_seq['sequences']
        ref = compress_seq['reference']
        is_vcf = True
    else:
        aln = args.alignment

    tt = ancestral_sequence_inference(tree=T, aln=aln, ref=ref, marginal=args.inference)

    if is_vcf:
        # TreeTime overwrites ambig sites on tips during ancestral reconst.
        # Put these back in tip sequences now, to avoid misleading
        tt.recover_var_ambigs()

    anc_seqs['nodes'] = collect_sequences_and_mutations(T, is_vcf)

    if args.output:
        anc_seqs_fname = args.output
    else:
        anc_seqs_fname = '.'.join(args.alignment.split('.')[:-1]) + '.anc_seqs.json'

    anc_seqs_success = write_json(anc_seqs, anc_seqs_fname)
    print("ancestral sequences written to",anc_seqs_fname, file=sys.stdout)

    # If VCF, output VCF including new ancestral seqs
    if is_vcf:
        if args.output_vcf:
            vcf_fname = args.output_vcf
        else:
            vcf_fname = '.'.join(args.alignment.split('.')[:-1]) + '.vcf'
        write_vcf(tt.get_tree_dict(keep_var_ambigs=True), vcf_fname)
        print("ancestral sequences as vcf-file written to",vcf_fname, file=sys.stdout)

    if anc_seqs_success:
        return 0
    else:
        return 1



"""Calculate the distance between amino acid sequences across entire genes or at a predefined subset of sites.
"""
import argparse
import Bio
import Bio.Phylo
from collections import defaultdict
import json
import numpy as np
import os
import sys

from .reconstruct_sequences import load_alignments


def read_masks(mask_file):
    ha_masks = {}
    with open(mask_file) as f:
        for line in f:
            (key, value) = line.strip().split()
            ha_masks[key] = np.frombuffer(value.encode(), 'S1').astype(int).astype(bool)

    return ha_masks


def mask_sites(aa, mask):
    return aa[mask[:len(aa)]]


def mask_distance(aaA, aaB, mask):
    """Return distance of sequences aaA and aaB by comparing sites in the given binary mask.

    >>> aaA = np.array(["A", "B", "C"], dtype="S1")
    >>> aaB = np.array(["A", "B", "D"], dtype="S1")
    >>> mask = np.array([0, 1, 1], dtype=np.bool)
    >>> mask_distance(aaA, aaB, mask)
    1
    >>> aaB = np.array(["A", "B", "X"], dtype="S1")
    >>> mask_distance(aaA, aaB, mask)
    0
    """
    sites_A = mask_sites(aaA, mask)
    sites_B = mask_sites(aaB, mask)

    # Count sites that differ between sequences excluding undetermined residues.
    distance = int(np.sum((sites_A != sites_B) & (sites_A != b"X") & (sites_B != b"X")))

    return distance


def register_arguments(parser):
    parser.add_argument("--tree", help="Newick tree", required=True)
    parser.add_argument("--alignment", nargs="+", help="sequence(s) to be used, supplied as FASTA files", required=True)
    parser.add_argument('--gene-names', nargs="+", type=str, help="names of the sequences in the alignment, same order assumed", required=True)
    parser.add_argument("--output", help="JSON file with calculated distances stored by node name and attribute name", required=True)
    parser.add_argument("--attribute-names", nargs="+", help="names to store distances associated with the corresponding masks", required=True)
    parser.add_argument("--masks", help="tab-delimited mask definitions with mask name in first column and binary mask in second column")
    parser.add_argument("--mask-names", nargs="*", help="name of each mask to use from the given masks file. Distances are calculated for each given mask and stored in each of the corresponding attribute names. If no mask is provided, all sites will be used.")


def run(args):
    # Load tree.
    tree = Bio.Phylo.read(args.tree, "newick")

    # Identify the name of the root node.
    root_node_name = tree.root.name

    # Load sequences.
    alignments = load_alignments(args.alignment, args.gene_names)

    # Index sequences by node name and gene.
    sequences_by_node_and_gene = defaultdict(dict)
    for gene, alignment in alignments.items():
        for record in alignment:
            sequences_by_node_and_gene[record.name][gene] = str(record.seq)

    # Create a single amino acid sequence for the root node based on the order of the genes.
    # Convert the string into an array for mask comparisons.
    root_node_sequence = np.fromstring(
        "".join([
            sequences_by_node_and_gene[root_node_name][gene]
            for gene in args.gene_names
        ]),
        dtype="S1"
    )

    # Determine which sites to include in distance calculations.
    # If any masks are specified, calculate the distance for each mask and store it in the corresponding attribute name.
    if args.masks:
        # Load masks.
        masks = read_masks(args.masks)

        # Map masks to attribute names where distances will be stored.
        attributes_by_mask = dict(zip(args.mask_names, args.attribute_names))
    else:
        # If no masks are specified, calculate the distance using all sites.
        mask_name = "all_sites"
        masks = {mask_name: np.ones(len(root_node_sequence)).astype(bool)}

        # Use the requested attribute name to store distances for this mask.
        attributes_by_mask = {mask_name: args.attribute_names[0]}

    # Calculate Hamming distance between the root node's sequence and the single
    # amino acid sequence for each node in the tree.
    distances_by_node = {}
    for node, node_translations in sequences_by_node_and_gene.items():
        node_sequence = np.fromstring(
            "".join([
                node_translations[gene]
                for gene in args.gene_names
            ]),
            dtype="S1"
        )

        # Calculate the distance for this node for each requested mask and store
        # the result in the corresponding attribute name.
        distances_by_node[node] = {}
        for mask_name, attribute in attributes_by_mask.items():
            distances_by_node[node][attribute] = mask_distance(
                root_node_sequence,
                node_sequence,
                masks[mask_name]
            )

    # Export distances to JSON.
    with open(args.output, "w") as oh:
        json.dump({"nodes": distances_by_node}, oh, indent=1, sort_keys=True)



"""
Assign clades to nodes in a tree based on amino-acid or nucleotide signatures.
"""

import sys
from Bio import Phylo
import pandas as pd
from .utils import get_parent_name_by_child_name_for_tree, read_node_data, write_json

def read_in_clade_definitions(clade_file):
    '''
    Reads in tab-seperated file that defines clades by amino acid or nucleotide mutations

    Format:
    clade	gene	site alt
    Clade_1	ctpE    81  D
    Clade_2	nuc 30642   T
    Clade_3	nuc 444296  A
    Clade_4	pks8    634 T
    '''

    clades = {}

    df = pd.read_csv(clade_file, sep='\t' if clade_file.endswith('.tsv') else ',')
    for index, row in df.iterrows():
        allele = (row.gene, row.site, row.alt)
        if row.clade in clades:
            clades[row.clade].append(allele)
        else:
            clades[row.clade] = [allele]

    return clades

def get_node_alleles(node_name, muts, parents):
    '''
    Retrieve the set of mutations leading to a node
    This includes node-specific mutations as well as mutations
    along its full line of descent
    '''
    node_alleles = []
    sites_encountered = set()
    focal_node = node_name
    while focal_node is not None:
        if 'aa_muts' in muts[focal_node]:
            for gene in muts[focal_node]['aa_muts']:
                for mut in muts[focal_node]['aa_muts'][gene]:
                    site = (gene, int(mut[1:-1]))
                    if site not in sites_encountered:
                        sites_encountered.add(site)
                        allele = (gene, int(mut[1:-1]), mut[-1])
                        node_alleles.append(allele)
        if 'muts' in muts[focal_node]:
            for mut in muts[focal_node]['muts']:
                site = (gene, int(mut[1:-1]))
                if site not in sites_encountered:
                    sites_encountered.add(site)
                    allele = ('nuc', int(mut[1:-1]), mut[-1])
                    node_alleles.append(allele)

        focal_node = parents.get(focal_node)

    return node_alleles

def is_node_in_clade(clade_alleles, node_alleles):
    '''
    Determines whether a node contains all mutations that define a clade
    '''
    is_clade = False
    # mutations are stored on nodes in format 'R927H' matching the clade definitions
    # if all of the clade-defining mutations are in the node mutations, it's part of this clade
    if all([ clade_allele in node_alleles for clade_allele in clade_alleles ]):
        is_clade = True
    return is_clade

def assign_clades(clade_designations, all_muts, tree):
    '''
    Ensures all nodes have an entry (or auspice doesn't display nicely), tests each node
    to see if it's the first member of a clade (assigns 'clade_annotation'), and sets
    all nodes's clade_membership to the value of their parent. This will change if later found to be
    the first member of a clade.
    '''

    clade_membership = {}
    parents = get_parent_name_by_child_name_for_tree(tree)

    # first pass to set all nodes to unassigned as precaution to ensure attribute is set
    for node in tree.find_clades(order = 'preorder'):
        clade_membership[node.name] = {'clade_membership': 'unassigned'}

    # second pass to assign 'clade_annotation' to basal nodes within each clade
    # if multiple nodes match, assign annotation to largest
    # otherwise occasional unwanted cousin nodes get assigned the annotation
    for clade_name, clade_alleles in clade_designations.items():
        node_counts = {}
        for node in tree.find_clades(order = 'preorder'):
            node_alleles = get_node_alleles(node.name, all_muts, parents)
            if is_node_in_clade(clade_alleles, node_alleles):
                node_counts[node.name] = node.count_terminals()
        sorted_nodes = list(sorted(node_counts.items(), key=lambda x: x[1], reverse=True))
        if len(sorted_nodes) > 0:
            target_node = sorted_nodes[0][0]
            clade_membership[target_node] = {'clade_annotation': clade_name, 'clade_membership': clade_name}

    # third pass to propagate 'clade_membership'
    # don't propagate if encountering 'clade_annotation'
    for node in tree.find_clades(order = 'preorder'):
        for child in node:
            if 'clade_annotation' not in clade_membership[child.name]:
                clade_membership[child.name]['clade_membership'] = clade_membership[node.name]['clade_membership']

    return clade_membership


def register_arguments(parser):
    parser.add_argument('--tree', help="prebuilt Newick -- no tree will be built if provided")
    parser.add_argument('--mutations', nargs='+', help='JSON(s) containing ancestral and tip nucleotide and/or amino-acid mutations ')
    parser.add_argument('--clades', type=str, help='TSV file containing clade definitions by amino-acid')
    parser.add_argument('--output', type=str, help="name of JSON files for clades")


def run(args):
    ## read tree and data, if reading data fails, return with error code
    tree = Phylo.read(args.tree, 'newick')
    node_data = read_node_data(args.mutations, args.tree)
    if node_data is None:
        print("ERROR: could not read node data (incl sequences)")
        return 1
    all_muts = node_data['nodes']

    clade_designations = read_in_clade_definitions(args.clades)

    clade_membership = assign_clades(clade_designations, all_muts, tree)

    write_json({'nodes': clade_membership}, args.output)
    print("clades written to", args.output, file=sys.stdout)



from setuptools import setup
from pathlib    import Path

base_dir     = Path(__file__).parent.resolve()
version_file = base_dir / "augur/__version__.py"
readme_file  = base_dir / "README.md"

# Eval the version file to get __version__; avoids importing our own package
with version_file.open() as f:
    exec(f.read())

# Get the long description from the README file
with readme_file.open(encoding = "utf-8") as f:
    long_description = f.read()

setup(
        name = "nextstrain-augur",
        version = __version__,
        author = "Nextstrain developers",
        author_email = "trevor@bedford.io, richard.neher@unibas.ch",
        description = "A bioinformatics toolkit for phylogenetic analysis.",
        long_description = long_description,
        long_description_content_type = "text/markdown",
        keywords = "nextstrain, molecular epidemiology",
        url = "https://github.com/nextstrain/augur",
        project_urls = {
            "Bug Reports": "https://github.com/nextstrain/augur/issues",
            "Change Log":  "https://github.com/nextstrain/augur/blob/master/CHANGES.md#next",
            "Source":      "https://github.com/nextstrain/augur",
        },
        packages=['augur'],
        package_data={'augur': ['data/*']},
        data_files = [("", ["LICENSE.txt"])],
        python_requires = '>=3.4',
        install_requires = [
            "bcbio-gff >=0.6.4, ==0.6.*",
            "biopython >=1.73, ==1.*",
            "boto >=2.38, ==2.*",
            "cvxopt >=1.1.9, ==1.1.*",
            "ipdb >=0.10.1",
            "jsonschema ==3.0.0a3",
            "matplotlib >=2.0, ==2.*",
            "pandas >=0.23.4, ==0.23.*",
            "phylo-treetime >=0.5.2, ==0.5.*",
            "seaborn >=0.9.0, ==0.9.*",
            "snakemake >=5.1.5, ==5.*"
        ],
        extras_require={
            'dev': [
                "pylint >=1.7.6, ==1.7.*",
                "pytest >=3.2.1, ==3.*",
                "wheel >=0.32.3, ==0.32.*"
            ]
        },
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "License :: OSI Approved :: GNU Affero General Public License v3",

            # Python 3 only; pathlib is >=3.4
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            ],

        # Install an "augur" program which calls augur.__main__.main()
        #   https://setuptools.readthedocs.io/en/latest/setuptools.html#automatic-script-creation
        entry_points = {
            "console_scripts": [
                "augur = augur.__main__:main",
            ],
        },
        )
